#     pbart.R
#     Copyright (C) 2012-2016 Matthew T. Pratola and Robert E. McCulloch
#
#     This file is part of MPIBART.
#
#     MPIBART is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     MPIBART is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#     Author contact information
#     Matthew T. Pratola: mpratola@gmail.com

#--------------------------------------------------
serBart=function(cstem,
  x,y,xp=NULL,
  nd=1000,burn=500,
  m=200,
  sigest=NA, sigdf=3, sigquant=.90,
  kfac=2.0,
  fname="")
{
  cat("***** Running serBart\n\n")
  my=mean(y)
  y=y-my
  #--------------------------------------------------
  #write files
  xroot=paste(fname,"x.txt",sep="")
  yroot=paste(fname,"y.txt",sep="")
  xproot=paste(fname,"xp.txt",sep="")

  p=ncol(x)
  write(t(x),file=xroot,ncol=p)
  write(y,file=yroot,ncol=1)
  write(t(xp),file=xproot,ncol=p)
  #--------------------------------------------------
  #set BART parameters
  if(is.na(sigest)) {
    df = data.frame(x,y)
    lmf = lm(y~.,df)
    sighat = summary(lmf)$sigma
  } else {
    sighat = sigest
  }
  cat('sighat: ',sighat,'\n')
  nu=sigdf
  sigq=qchisq(1.0-sigquant,nu)
  lambda = (sighat*sighat*sigq)/nu #lambda parameter for sigma prior
  #--------------------------------------------------
  #run bart
  cmd = paste(cstem,'bart',sep='')
  cmd = paste(cmd,xroot,yroot,xproot,nd,lambda,burn,m,nu,kfac)
  cat(cmd,'\n')
  system(cmd)
#--------------------------------------------------
#read in results
#-rw-rw-r-- 1 rem rem 833K Jun 20 17:57 bart-pred-fit.txt
#-rw-rw-r-- 1 rem rem 124K Jun 20 17:57 bart-fit.txt
#-rw-rw-r-- 1 rem rem    4 Jun 20 17:57 time.txt
#-rw-rw-r-- 1 rem rem  12K Jun 20 17:57 bart-sd.txt
  sigma=scan('bart-sd.txt')
  yhattrain = scan('bart-fit.txt')
  yhattest = scan('bart-pred-fit.txt')
  timetorun = scan('time.txt')
  ats = scan('bart-ave-tree-size.txt')
  anb = scan('bart-ave-num-bots.txt')
  #--------------------------------------------------
  #return
  retlist=list(sigma=sigma,yhattrain=yhattrain+my,timetorun=timetorun,
    averagetreesize=ats,averagenumbots=anb)
  if(!is.null(xp)) retlist$yhattest=yhattest+my
  return(retlist)
}
#--------------------------------------------------
#--------------------------------------------------
mpiBart=function(numnodes,cstem,
  x,y,xp=NULL,
  nd=1000,burn=500,
  m=200,
  sigest=NA, sigdf=3, sigquant=.90,
  kfac=2.0,
  fname="",
  thin=1,type='p',
  numnodespred=numnodes,pause=FALSE,verbose=FALSE)
{
  if (verbose) cat("***** Running mpiBart\n\n")
  my=mean(y)
  y=y-my
  if(!is.null(xp) && nrow(xp)<2) {
    xp=NULL
    cat('@@@@@@@@@@@@@ prediction x has less than two observations, set to NULL\n')
  }
  if(!is.null(xp)) {
     if(floor(nrow(xp)/numnodespred) ==0) {
        numnodespred=2
        cat('@@@@@@@@@@@@@@@@@@@@warning: numnodespred set to 2\n')
     }
  }
  #--------------------------------------------------
  #write files
  xroot=paste(fname,"x",sep="")
  yroot=paste(fname,"y",sep="")
  xproot=paste(fname,"xp",sep="")
  if (verbose){
    cat('xroot,yroot,xproot:\n')
    print(xroot);print(yroot);print(xproot)
  }
  tempfiles = writeBartMpiFiles(numnodes,numnodespred,x,y,xp,
    xroot,yroot,xproot,verbose=verbose)
  # above command saves names of temp files for later deletion
  #--------------------------------------------------
  #set BART parameters
  if(is.na(sigest)) {
    df = data.frame(x,y)
    lmf = lm(y~.,df)
    sighat = summary(lmf)$sigma
  } else {
    sighat = sigest
  }
  if (verbose) cat('sighat: ',sighat,'\n')
  nu=sigdf
  sigq=qchisq(1.0-sigquant,nu)
  lambda = (sighat*sighat*sigq)/nu #lambda parameter for sigma prior
  #--------------------------------------------------
  #run mpibart
  if (verbose) cat('*****running mcmc code in pbart.R\n')
  cmd = paste('mpirun -np ',numnodes,' ',cstem,'mpibart',sep='')
  cmd = paste(cmd,xroot,yroot)
  cmd = paste(cmd,thin,nd,lambda,burn,m,nu,kfac,fname)
  if (verbose) {
    cat('cmd:\n')
    cat(cmd)
    cat('\n')
  }
  if (pause) readline("hit <enter> to continue")
  ptime = system.time({system(cmd)})
  #--------------------------------------------------
  #read in results from mpibart
  yhtrain = getBartMpiTrainFits(numnodes,fname,verbose=verbose)
  sddr = scan(paste(fname,'bart-sd.txt',sep=''),quiet=!verbose)
  timetorun = scan(paste(fname,'time.txt',sep=''),quiet=!verbose)
  #--------------------------------------------------
  #run prediction code
  if(!is.null(xp)) {
    if (verbose) cat('*****running prediction code in pbart.R\n')
    cmd = paste('mpirun -np ',numnodespred,' ',cstem, 'bartpred ',
      xproot,' ',type,' ',fname,sep='') 
    if (verbose){
      cat('cmd:\n')
      cat(cmd)
      cat('\n')
    }
    if (pause) readline("hit <enter> to continue")
    system(cmd)
    #--------------------------------------------------
    #get prediction ouput
    if(!is.null(xp)) {
      outsamp = getBartMpiTestFits(numnodespred,type,fname,verbose=verbose)
    }
  }
  #--------------------------------------------------
  #return
  retlist = list(yhattrain=yhtrain+my,sigma=sddr)
  if(!is.null(xp)) {
    retlist$yhattest = outsamp$yhat+my
    retlist$ysdtest = outsamp$yhatsd
  } else {
    retlist$yhattest = NULL
    retlist$ysdtest = NULL
  }
  retlist$timetorun=timetorun
  retlist$my = my
  tempfiles = c(tempfiles,paste(fname,'bart-sd.txt',sep=""))

  system(paste(c('rm',tempfiles),collapse=' '))
  return(retlist)
}
#--------------------------------------------------
mpiPred = function(retlist,numnodespred,cstem,xp,type='m',fname='',
  verbose=FALSE)
{
  # function that writes out necessary test set files plus junk for
  # training, then calls mpipred.  We assume that "trees.txt" has the
  # trees for which predictions are desired.
  xroot=paste(fname,"x",sep="")
  yroot=paste(fname,"y",sep="")
  xproot=paste(fname,"xp",sep="")

  tempfiles = writeBartMpiFiles(numnodespred,numnodespred,
     x=matrix(0,nrow=100,ncol=ncol(xp)),y=rep(0,100),xp,
     xroot,yroot,xproot,verbose=verbose)
  # above writes data + records temp filenames for later deletion
  cmd = paste('mpirun -np ',numnodespred,' ',cstem, 'bartpred ',
    xproot,' ',type,' ',fname,sep='') 
  system(cmd)
  #--------------------------------------------------
  #get prediction ouput
  if(!is.null(xp)) {
    outsamp = getBartMpiTestFits(numnodespred,type,fname,verbose=verbose)
  }
  outsamp$yhat = outsamp$yhat + retlist$my
  system(paste(c('rm',tempfiles),collapse=' '))
  return(outsamp)
}

#--------------------------------------------------
getBartMpiTestFits=function(numnodespred,type='p',fname='',verbose=FALSE) {
  tempfiles = NULL # list of files to delete after reading
  if (type=='m') {
    for(i in 1:numnodespred) {
      fnm = paste(fname,'ypost',i-1,'.txt',sep='')
      tempfiles = c(tempfiles,fnm)
      temp = read.table(fnm,colClasse='numeric')
      if (i==1) yhat = t(temp) # t coerces data frame to matrix
      else yhat = rbind(yhat,t(temp))
    }
    system(paste(c('rm',tempfiles),collapse=' '))
    return(list(yhat=yhat))
  } else {
    bfit=c()
    bsd=c()
    for(i in 1:numnodespred) {
      fnm = paste(fname,'yp',i-1,'.txt',sep='')
      fsd = paste(fname,'ysd',i-1,'.txt',sep='')
      tempfiles = c(tempfiles,fnm,fsd)
      temp = scan(fnm,quiet=!verbose)
      if (verbose) print(length(temp))
      bfit = c(bfit,scan(fnm,quiet=!verbose))
      temp = scan(fsd,quiet=!verbose)
      if (verbose) print(length(temp))
      bsd = c(bsd,scan(fsd),quiet=!verbose)
    }
    system(paste(c('rm',tempfiles),collapse=' '))
    return(list(yhat=bfit,yhatsd=bsd))
  }
}
#--------------------------------------------------
#--------------------------------------------------
getBartMpiTrainFits=function(numnodes,fname='',verbose=FALSE) {
  tempfiles = NULL # list of files to delete after reading
  bfit=c()
  for(i in 1:(numnodes-1)) {
    fnm = paste(fname,'bart-fit',i,'.txt',sep='')
    tempfiles = c(tempfiles,fnm)
    bfit = c(bfit,scan(fnm,quiet=!verbose))
  }
  system(paste(c('rm',tempfiles),collapse=' '))
  return(bfit)
}
#--------------------------------------------------
#--------------------------------------------------
writeBartMpiFiles=function(numnodes,numnodespred,x,y,xp=NULL,xroot="x",yroot="y",xproot="xp",verbose=FALSE) {

  filelist = NULL # We'll return list of files created on function exit
  p=ncol(x)
  n=nrow(x)

  pv=rep(0,numnodes-1) #number of in-sample observations for each slave node
  pv[1:(numnodes-2)] = floor(n/(numnodes-1)) #number of slave = numnodes-1
  pv[numnodes-1]=n-sum(pv)

  if (verbose) cat("Generating ",numnodes-1," train datasets of sizes ",pv,"\n")
  for(i in 1:(numnodes-1)) #c indices of slaves
  {
    if(i==1) {
      ii=1:pv[1]
    } else {
      ii=(sum(pv[1:(i-1)])+1):sum(pv[1:i])
    }

    xx=x[ii,]
    yy=y[ii]
    tempfile = paste(c(yroot,xroot),i,'.txt',sep="") 
    write(yy,ncol=1,file=tempfile[1])
    write(t(xx),ncol=p,file=tempfile[2])
    filelist = c(filelist,tempfile)
  }

  if(!is.null(xp)) {
    pv=rep(0,numnodespred) #number of out-of-sample observations for each slave node
    pv[1:(numnodespred-1)] = floor(nrow(xp)/numnodespred) #number of slaves = numnodespred
    pv[numnodespred]=nrow(xp)-sum(pv)

    if (verbose) cat("Generating ",numnodespred," predictive X datasets of sizes ",pv,"\n")
    for(i in 1:numnodespred) #note xpi.txt i=1,2,..numnodespred 
    {
      if(i==1) { 
        ii=1:pv[1]
      } else {
        ii=(sum(pv[1:(i-1)])+1):sum(pv[1:i])
      }
        
      xx=xp[ii,]
      #yy=fp[ii]
      tempfile = paste(xproot,i,'.txt',sep="")
      write(t(xx),ncol=p,file=tempfile)
      filelist = c(filelist,tempfile)
    }
  }
  return(filelist)
}
